/*
 * GroupLens Common Utilities
 * Copyright © 2012 Regents of the University of Minnesota
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *  * Neither the name of the University of Minnesota nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software has been partly funded by NSF grant IIS 08-08692.
 */
package org.grouplens.common.dto;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.grouplens.common.dto.Dto.Exclude;
import org.grouplens.common.dto.Dto.ExcludeType;
import org.grouplens.common.dto.Dto.Named;
import org.grouplens.common.dto.Dto.XmlAttribute;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

public class XmlDtoContentHandler implements DtoContentHandler {

	private final boolean ignoreExclude;
	private final boolean prettyPrint;
	private static final int TAB_WIDTH = 4;

	public XmlDtoContentHandler() {
		this(false, true);
	}

	public XmlDtoContentHandler(boolean ignoreExclude) {
		this(ignoreExclude, true);
	}

	public XmlDtoContentHandler(boolean ignoreExclude, boolean prettyPrint) {
		this.ignoreExclude = ignoreExclude;
		this.prettyPrint = prettyPrint;
	}

	@Override
	public <T extends Dto> String toString(DtoContainer<T> dto) {
		StringWriter writer = new StringWriter();
		this.toString(dto, writer);
		return writer.toString();
	}

	@Override
	public <T extends Dto> void toString(DtoContainer<T> dto, Writer writer) {
		try {
			Document xml = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().newDocument();
			xml.setXmlStandalone(true);
			Element root;

			if (dto.isMultiValued()) {
				root = xml.createElement(getDtoPluralName(dto.getDtoType()));
				for (T instance : dto.get()) {
					root.appendChild(serializeDto(instance,
						getDtoSingularName(instance.getClass()), xml));
				}
			} else {
				root = serializeDto(dto.getSingle(), getDtoSingularName(dto.getDtoType()), xml);
			}
			xml.appendChild(root);

			TransformerFactory tf = TransformerFactory.newInstance();
			tf.setAttribute("indent-number", TAB_WIDTH);
			Transformer trans = tf.newTransformer();
			if (prettyPrint) {
				trans.setOutputProperty(OutputKeys.INDENT, "yes");
			}
			trans.transform(new DOMSource(xml), new StreamResult(writer));

		} catch (ParseException pe) {
			// we allow the try body to throw parse Exceptions
			throw pe;
		} catch (Exception e) {
			// Any other exception is treated as the root cause of a parse failure.
			throw new ParseException(e);
		}
	}

	private <T> Element serializeDto(T instance, String nodeName, Document xml) {
		Element root = xml.createElement(nodeName);
		
		for (Field f : instance.getClass().getFields()) {
			if (!shouldSkipField(f, false)) {
				Object fieldInstance;
				try {
					fieldInstance = f.get(instance);
				} catch (IllegalAccessException e) {
					// just skip the field if it can't be accessed
					continue;
				}

				if (fieldInstance != null) {
					if (Dto.class.isAssignableFrom(f.getType())) {
						root.appendChild(serializeDto(fieldInstance, translateFieldName(f), xml));
					} else if (f.getType().isArray()) {
						root.appendChild(serializeArray(fieldInstance, f, xml));
					} else {
						String tagName = translateFieldName(f);
						String value = escapeString(fieldInstance.toString());
						if (f.getAnnotation(XmlAttribute.class) != null) {
							root.setAttribute(tagName, value);
						} else {
							Element e = xml.createElement(tagName);
							e.setTextContent(value);
							root.appendChild(e);
						}
					}
				}
			}
		}
		return root;
	}

	private Element serializeArray(Object array, Field f, Document xml) {
		Class<?> elementType = f.getType().getComponentType();
		Element root = xml.createElement(translateFieldName(f));

		for (int i = 0; i < Array.getLength(array); i++) {
			Object currInstance = Array.get(array, i);
			Element nextChild;
			if (currInstance == null) {
				nextChild = xml.createElement("null");
			} else if (Dto.class.isAssignableFrom(elementType)) {
				@SuppressWarnings("unchecked")
				String tagName = getDtoSingularName((Class<? extends Dto>)elementType);
				nextChild = serializeDto(currInstance, tagName, xml);
			} else {
				nextChild = xml.createElement(elementType.getSimpleName());
				String content = escapeString(currInstance.toString());
				nextChild.setTextContent(content);
			}
			root.appendChild(nextChild);
		}		
		return root;
	}

	@Override
	public <T extends Dto> void fromString(String content, DtoContainer<T> container) {
		// If the body is empty, we leave the container as empty
		if (!content.trim().isEmpty()) {
			try {
				Document xml = DocumentBuilderFactory.newInstance().newDocumentBuilder().
					parse(new InputSource(new StringReader(content)));
				String rootTagName = xml.getDocumentElement().getTagName();
				
				if (container.isMultiValued()) {
					String expectedTagName = getDtoPluralName(container.getDtoType());
					if (!rootTagName.equals(expectedTagName)) {
						throw new ParseException("Expected " + expectedTagName + 
							" as root tag but found " + rootTagName);
					}
					
					List<T> dtos = new ArrayList<T>();
					for (Element currElem : getChildElements(xml.getDocumentElement())) {
						dtos.add(deserializeDto(container.getDtoType(), currElem));
					}					
					container.set(dtos);					
				} else {				
					String expectedTagName = getDtoSingularName(container.getDtoType());
					if (!rootTagName.equals(expectedTagName)) {
						throw new ParseException("Expected " + expectedTagName + 
							" as root tag but found " + rootTagName);
					}					
					container.set((T)deserializeDto(container.getDtoType(),
						xml.getDocumentElement()));
				}
				
			} catch (ParseException pe) {
				// we allow the try body to throw parse Exceptions
				throw pe;
			} catch (Exception e) {
				// Any other exception is treated as the root cause of a parse failure.
				throw new ParseException(e);
			}
		}
	}

	@Override
	public <T extends Dto> void fromString(Reader reader, DtoContainer<T> container) {
		try {
			// Dump contents of the reader into a string and delegate so we can
			// easily check for an empty body
			fromString(dumpContents(reader), container);
		} catch (IOException e) {
			throw new ParseException(e);
		}
	}
	
	@SuppressWarnings("unchecked")
	private <T> T deserializeSimpleObject(Class<T> type, String content) {
		if (type.equals(String.class)) {
			return (T)unescapeString(content);
		} else if (type.equals(Character.class)) {
			if (content.trim().isEmpty()) {
				return (T)new Character('\0');
			} else if (content.trim().length() > 1) {
				throw new ParseException("Expected single character but found " + content);
			} else {
				return (T)new Character(content.charAt(0));
			}
		} else if (type.equals(Boolean.class)) {
			// Java's valueOf method does not reject improper Strings
			return (T)strictBooleanValueOf(content);
		} else if (!content.isEmpty()) {
			try {
				// Last resort is to attempt to invoke a "valueOf" method
				Method valueOf = type.getMethod("valueOf", String.class);
				return (T)valueOf.invoke(null, content);
			} catch (NoSuchMethodException nsme) {
				throw new ParseException("Unable to parse object of type " + 
					type.getName(), nsme);
			} catch (InvocationTargetException ite) {
				throw new ParseException("The text " + content + " does not define a valid " + 
					type.getName(), ite);
			} catch (IllegalAccessException iae) {
				throw new ParseException("Unable to access valueOf method to parse type " +
					type.getName(), iae);
			}
		} else {
			return null;
		}
	}

	@SuppressWarnings("unchecked")
	private <T extends Dto> T deserializeDto(Class<T> dtoType, Element elem) {
		if (getChildElements(elem).isEmpty() && !elem.getTextContent().trim().isEmpty()) {
			throw new ParseException("The text " + elem.getTextContent() + 
				" does not define a valid " + getDtoSingularName(dtoType));
		}

		T instance;
		try {
			instance = dtoType.newInstance();
		} catch (Exception e) {
			throw new ParseException("Could not instantiate object of type " +
				dtoType.getName(), e);
		}

		for (Field f : dtoType.getFields()) {
			if (!shouldSkipField(f, true)) {
				String fieldName = translateFieldName(f);
				List<Element> matchingTags = getTopLevelElementsByTagName(elem, fieldName);
				if (matchingTags.size() > 1) {
					throw new ParseException("Multiple tag pairs found for field " + f.getName());
				} else if (matchingTags.isEmpty()) {
					// Check if the field is serialized as an attribute rather than a tag pair
					Node attr = elem.getAttributes().getNamedItem(fieldName);
					if (attr != null) {
						setField(f, instance, deserializeSimpleObject(f.getType(),
							attr.getTextContent()));
					}
				} else if (Dto.class.isAssignableFrom(f.getType())) {
					setField(f, instance, deserializeDto((Class<? extends Dto>)f.getType(),
						matchingTags.get(0)));
				} else if (f.getType().isArray()) {
					Element root = matchingTags.get(0);
					setField(f, instance, deserializeArray(f.getType().getComponentType(),
						getChildElements(root).size(), root));
				} else {
					String content = matchingTags.get(0).getTextContent();
					setField(f, instance, deserializeSimpleObject(f.getType(), content));
				}
			}
		}
		return instance;
	}

	@SuppressWarnings("unchecked")
	private <T> T[] deserializeArray(Class<T> componentType, int length, Element elem) {
		T[] arr = (T[])Array.newInstance(componentType, length);
		List<Element> childElements = getChildElements(elem);
		
		if (childElements.isEmpty() && !elem.getTextContent().trim().isEmpty()) {
			throw new ParseException("Proper array is not defined by " + elem.getTextContent());
		}
		
		int i = 0;
		for (Element currElem : childElements) {
			String tagName = currElem.getTagName();
			if (tagName.equals("null")) {
				// skip the null element
				i++;
			} else if (Dto.class.isAssignableFrom(componentType)) {
				Class<? extends Dto> dtoType = (Class<? extends Dto>)componentType;
				String expectedTagName = getDtoSingularName(dtoType);
				if (!tagName.equals(expectedTagName)) {
					throw new ParseException("Expected " + expectedTagName + 
						" as tag name but found " + tagName);
				} else {
					arr[i++] = (T)deserializeDto(dtoType, currElem);
				}
			} else {
				String expectedTagName = componentType.getSimpleName();
				if (!tagName.equals(expectedTagName)) {
					throw new ParseException("Expected " + expectedTagName +
						" as tag name but found " + tagName);
				} else {
					arr[i++] = deserializeSimpleObject(componentType, currElem.getTextContent());
				}			
			}
		}
		return arr;
	}

	private boolean shouldSkipField(Field f, boolean forInput) {
		// we add the DtoContainer because we want to allow root DtoContainers
		// to be serialized, but not ones that are used as fields
		if (!Modifier.isPublic(f.getModifiers())
			|| f.getType().isAssignableFrom(DtoContainer.class)) {
			return true;
		}

		if (!ignoreExclude) {
			Exclude exclude = f.getAnnotation(Exclude.class);
			// if exclude is neither EXPORT nor IMPORT, the field is always skipped
			// if exclude is EXPORT then input fields are still included
			// if exclude is IMPORT then output fields are still included
			if (exclude != null) {
				ExcludeType[] types = exclude.value();
				ExcludeType skipIfPresent = (forInput ? ExcludeType.IMPORT : ExcludeType.EXPORT);
				for (int i = 0; i < types.length; i++) {
					if (skipIfPresent.equals(types[i]))
						return true;
				}
			}
		}
		return false;
	}

	private static String translateFieldName(Field f) {
		Named named = f.getAnnotation(Named.class);
		if (named != null) {
			return f.getType().isArray() ? named.plural() : named.singular();
		} else {
			return f.getName();
		}
	}

	private static String getDtoSingularName(Class<? extends Dto> type) {
		Named name = type.getAnnotation(Named.class);
		if (name != null) {
			return name.singular();
		} else {
			// if the first letter is capitalized, make it lower case.
			String derived = type.getSimpleName();
			String suffix = "Dto";
			derived = Character.toLowerCase(derived.charAt(0)) + derived.substring(1);
	
			// if the name ends with suffix, remove the suffix.
			if (derived.endsWith(suffix)) {
				derived = derived.substring(0, derived.length() - suffix.length());
			}
			return derived;
		}
	}

	private static String getDtoPluralName(Class<? extends Dto> type) {
		Named name = type.getAnnotation(Named.class);
		if (name != null) {
			return name.plural();
		} else {
			// Dto default names always assumed to be singular
			return getDtoSingularName(type) + "s";
		}
	}

	// Retrieve all child elements of an XML node
	private static List<Element> getChildElements(Node n) {
		NodeList children = n.getChildNodes();
		List<Element> childElements = new ArrayList<Element>(children.getLength());
		for (int i = 0; i < n.getChildNodes().getLength(); i++) {
			Node current = children.item(i);
			if (current.getNodeType() == Node.ELEMENT_NODE) {
				childElements.add((Element) current);
			}
		}
		return childElements;
	}

	// Retrieve elements with a matching tag name, but only if they are exactly
	// one level below the root element
	private static List<Element> getTopLevelElementsByTagName(Element root, String tagName) {
		ArrayList<Element> matchingTags = new ArrayList<Element>();
		Node temp = root.getFirstChild();
		while (temp != null) {
			if (temp.getNodeType() == Node.ELEMENT_NODE
			    && temp.getNodeName().equals(tagName)) {
				matchingTags.add((Element) temp);
			}
			temp = temp.getNextSibling();
		}
		return matchingTags;
	}

	private static String escapeString(String str) {
		return str.replace("\\", "\\\\").replace("\"", "\\\"")
			.replace("\t", "\\t").replace("\b", "\\b").replace("\n", "\\n")
			.replace("\r", "\\r").replace("\f", "\\f").replace("\'", "\\\'");
	}

	private static String unescapeString(String str) {
		return str.replace("\\\\", "\\").replace("\\\"","\"")
			.replace("\\t", "\t").replace("\\b", "\b").replace("\\n", "\n")
			.replace("\\r", "\r").replace("\\f", "\f").replace("\\\'", "\'");
	}

	private static Boolean strictBooleanValueOf(String text) {
		if (text.equalsIgnoreCase("true")) {
			return Boolean.TRUE;
		} else if (text.equalsIgnoreCase("false")) {
			return Boolean.FALSE;
		} else {
			throw new ParseException("Expected boolean but found " + text);
		}
	}

	// Store all of the contents available to a reader in a String
	private static String dumpContents(Reader reader) throws IOException {
		BufferedReader br = new BufferedReader(reader);
		StringBuilder builder = new StringBuilder();
		String line;
		try {
			while ((line = br.readLine()) != null) {
				builder.append(line);
				builder.append("\n");
			}
			return builder.toString();
		} finally {
			br.close();
		}
	}
	
	// Set the value of a Field object, wrapping any thrown exceptions in a ParseException
	private static void setField(Field f, Object instance, Object value) {
		try {
			f.set(instance, value);
		} catch (IllegalAccessException e) {
			throw new ParseException(e);
		}
	}
}